/*
 * kgraft_patch_bsc1053150
 *
 * Fix for CVE-2017-12762, bsc#1053150
 *
 *  Upstream commit:
 *  9f5af546e6ac ("isdn/i4l: fix buffer overflow")
 *
 *  SLE12-SP2 and -SP3 commit:
 *  Fix arrived via 4.4.80, i.e. e195a265512b593008871b8057f4a23de1cf04e0
 *
 *  Copyright (c) 2017 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_ISDN_I4L)

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/isdn.h>
#include <linux/fs.h>

#if !IS_MODULE(CONFIG_ISDN_I4L)
#error "KGR patch supports only CONFIG_ISDN_I4L=m."
#endif

#if !IS_ENABLED(CONFIG_ISDN_PPP)
#error "KGR patch supports only CONFIG_ISDN_PPP=y."
#endif

#define KGR_PATCHED_MODULE "isdn"

static isdn_dev **kgr_dev;

static isdn_net_dev * (*kgr_isdn_net_findif)(char *name);
static char * (*kgr_isdn_net_new)(char *name, struct net_device *master);
static int (*kgr_isdn_net_getpeer)(isdn_net_ioctl_phone *phone,
				isdn_net_ioctl_phone __user *peer);
static int (*kgr_isdn_net_rm)(char *name);
static int (*kgr_isdn_net_setcfg)(isdn_net_ioctl_cfg *cfg);
static int (*kgr_isdn_net_getcfg)(isdn_net_ioctl_cfg *cfg);
static int (*kgr_isdn_net_addphone)(isdn_net_ioctl_phone *phone);
static int (*kgr_isdn_net_getphones)(isdn_net_ioctl_phone *phone,
				char __user *phones);
static int (*kgr_isdn_net_delphone)(isdn_net_ioctl_phone *phone);
static int (*kgr_isdn_net_force_dial)(char *name);
static int (*kgr_isdn_ppp_dial_slave)(char *name);
static int (*kgr_isdn_ppp_hangup_slave)(char *name);
static int (*kgr_isdn_net_force_hangup)(char *name);
static int (*kgr_isdn_command)(isdn_ctrl *cmd);
static int (*kgr_isdn_ppp_ioctl)(int min, struct file *file,
				unsigned int cmd, unsigned long arg);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "isdn:dev", (void *)&kgr_dev },
	{ "isdn:isdn_net_findif", (void *)&kgr_isdn_net_findif },
	{ "isdn:isdn_net_new", (void *)&kgr_isdn_net_new },
	{ "isdn:isdn_net_getpeer", (void *)&kgr_isdn_net_getpeer },
	{ "isdn:isdn_net_rm", (void *)&kgr_isdn_net_rm },
	{ "isdn:isdn_net_setcfg", (void *)&kgr_isdn_net_setcfg },
	{ "isdn:isdn_net_getcfg", (void *)&kgr_isdn_net_getcfg },
	{ "isdn:isdn_net_addphone", (void *)&kgr_isdn_net_addphone },
	{ "isdn:isdn_net_getphones", (void *)&kgr_isdn_net_getphones },
	{ "isdn:isdn_net_delphone", (void *)&kgr_isdn_net_delphone },
	{ "isdn:isdn_net_force_dial", (void *)&kgr_isdn_net_force_dial },
	{ "isdn:isdn_ppp_dial_slave", (void *)&kgr_isdn_ppp_dial_slave },
	{ "isdn:isdn_ppp_hangup_slave", (void *)&kgr_isdn_ppp_hangup_slave },
	{ "isdn:isdn_net_force_hangup", (void *)&kgr_isdn_net_force_hangup },
	{ "isdn:isdn_ppp_ioctl", (void *)&kgr_isdn_ppp_ioctl },
};

/* from drivers/isdn/i4l/isdn_common.c, inlined */
static inline int kgr_isdn_minor2drv(int minor)
{
	return ((*kgr_dev)->drvmap[minor]);
}

/* from drivers/isdn/i4l/isdn_net.c, inlined */
static __inline__ int kgr_isdn_net_device_started(isdn_net_dev *n)
{
	isdn_net_local *lp = n->local;
	struct net_device *dev;

	if (lp->master)
		dev = lp->master;
	else
		dev = n->dev;
	return netif_running(dev);
}

/* patched */
char * kgr_isdn_net_newslave(char *parm)
{
	char *p = strchr(parm, ',');
	isdn_net_dev *n;
	char newname[10];

	if (p) {
		/*
		 * Fix CVE-2017-12762
		 *  -4 lines, +3 lines
		 */
		/* Slave-Name MUST not be empty or overflow 'newname' */
		if (strscpy(newname, p + 1, sizeof(newname)) <= 0)
			return NULL;
		*p = 0;
		/* Master must already exist */
		if (!(n = kgr_isdn_net_findif(parm)))
			return NULL;
		/* Master must be a real interface, not a slave */
		if (n->local->master)
			return NULL;
		/* Master must not be started yet */
		if (kgr_isdn_net_device_started(n))
			return NULL;
		return (kgr_isdn_net_new(newname, n->dev));
	}
	return NULL;
}

/* patched */
int kgr_isdn_ioctl(struct file *file, uint cmd, ulong arg)
{
	uint minor = iminor(file_inode(file));
	isdn_ctrl c;
	int drvidx;
	int ret;
	int i;
	char __user *p;
	char *s;
	union iocpar {
		char name[10];
		char bname[22];
		isdn_ioctl_struct iocts;
		isdn_net_ioctl_phone phone;
		isdn_net_ioctl_cfg cfg;
	} iocpar;
	void __user *argp = (void __user *)arg;

#define name  iocpar.name
#define bname iocpar.bname
#define iocts iocpar.iocts
#define phone iocpar.phone
#define cfg   iocpar.cfg

	if (minor == ISDN_MINOR_STATUS) {
		switch (cmd) {
		case IIOCGETDVR:
			return (TTY_DV +
				(NET_DV << 8) +
				(INF_DV << 16));
		case IIOCGETCPS:
			if (arg) {
				ulong __user *p = argp;
				int i;
				if (!access_ok(VERIFY_WRITE, p,
					       sizeof(ulong) * ISDN_MAX_CHANNELS * 2))
					return -EFAULT;
				for (i = 0; i < ISDN_MAX_CHANNELS; i++) {
					put_user((*kgr_dev)->ibytes[i], p++);
					put_user((*kgr_dev)->obytes[i], p++);
				}
				return 0;
			} else
				return -EINVAL;
			break;
		case IIOCNETGPN:
			/* Get peer phone number of a connected
			 * isdn network interface */
			if (arg) {
				if (copy_from_user(&phone, argp, sizeof(phone)))
					return -EFAULT;
				return kgr_isdn_net_getpeer(&phone, argp);
			} else
				return -EINVAL;
		default:
			return -EINVAL;
		}
	}
	if (!(*kgr_dev)->drivers)
		return -ENODEV;
	if (minor <= ISDN_MINOR_BMAX) {
		drvidx = kgr_isdn_minor2drv(minor);
		if (drvidx < 0)
			return -ENODEV;
		if (!((*kgr_dev)->drv[drvidx]->flags & DRV_FLAG_RUNNING))
			return -ENODEV;
		return 0;
	}
	if (minor <= ISDN_MINOR_CTRLMAX) {
/*
 * isdn net devices manage lots of configuration variables as linked lists.
 * Those lists must only be manipulated from user space. Some of the ioctl's
 * service routines access user space and are not atomic. Therefore, ioctl's
 * manipulating the lists and ioctl's sleeping while accessing the lists
 * are serialized by means of a semaphore.
 */
		switch (cmd) {
		case IIOCNETDWRSET:
			printk(KERN_INFO "INFO: ISDN_DW_ABC_EXTENSION not enabled\n");
			return (-EINVAL);
		case IIOCNETLCR:
			printk(KERN_INFO "INFO: ISDN_ABC_LCR_SUPPORT not enabled\n");
			return -ENODEV;
		case IIOCNETAIF:
			/* Add a network-interface */
			if (arg) {
				if (copy_from_user(name, argp, sizeof(name)))
					return -EFAULT;
				s = name;
			} else {
				s = NULL;
			}
			ret = mutex_lock_interruptible(&(*kgr_dev)->mtx);
			if (ret) return ret;
			if ((s = kgr_isdn_net_new(s, NULL))) {
				if (copy_to_user(argp, s, strlen(s) + 1)) {
					ret = -EFAULT;
				} else {
					ret = 0;
				}
			} else
				ret = -ENODEV;
			mutex_unlock(&(*kgr_dev)->mtx);
			return ret;
		case IIOCNETASL:
			/* Add a slave to a network-interface */
			if (arg) {
				if (copy_from_user(bname, argp, sizeof(bname) - 1))
					return -EFAULT;
				/*
				 * Fix CVE-2017-12762
				 * +1 line
				 */
				bname[sizeof(bname)-1] = 0;
			} else
				return -EINVAL;
			ret = mutex_lock_interruptible(&(*kgr_dev)->mtx);
			if (ret) return ret;
			if ((s = kgr_isdn_net_newslave(bname))) {
				if (copy_to_user(argp, s, strlen(s) + 1)) {
					ret = -EFAULT;
				} else {
					ret = 0;
				}
			} else
				ret = -ENODEV;
			mutex_unlock(&(*kgr_dev)->mtx);
			return ret;
		case IIOCNETDIF:
			/* Delete a network-interface */
			if (arg) {
				if (copy_from_user(name, argp, sizeof(name)))
					return -EFAULT;
				ret = mutex_lock_interruptible(&(*kgr_dev)->mtx);
				if (ret) return ret;
				ret = kgr_isdn_net_rm(name);
				mutex_unlock(&(*kgr_dev)->mtx);
				return ret;
			} else
				return -EINVAL;
		case IIOCNETSCF:
			/* Set configurable parameters of a network-interface */
			if (arg) {
				if (copy_from_user(&cfg, argp, sizeof(cfg)))
					return -EFAULT;
				return kgr_isdn_net_setcfg(&cfg);
			} else
				return -EINVAL;
		case IIOCNETGCF:
			/* Get configurable parameters of a network-interface */
			if (arg) {
				if (copy_from_user(&cfg, argp, sizeof(cfg)))
					return -EFAULT;
				if (!(ret = kgr_isdn_net_getcfg(&cfg))) {
					if (copy_to_user(argp, &cfg, sizeof(cfg)))
						return -EFAULT;
				}
				return ret;
			} else
				return -EINVAL;
		case IIOCNETANM:
			/* Add a phone-number to a network-interface */
			if (arg) {
				if (copy_from_user(&phone, argp, sizeof(phone)))
					return -EFAULT;
				ret = mutex_lock_interruptible(&(*kgr_dev)->mtx);
				if (ret) return ret;
				ret = kgr_isdn_net_addphone(&phone);
				mutex_unlock(&(*kgr_dev)->mtx);
				return ret;
			} else
				return -EINVAL;
		case IIOCNETGNM:
			/* Get list of phone-numbers of a network-interface */
			if (arg) {
				if (copy_from_user(&phone, argp, sizeof(phone)))
					return -EFAULT;
				ret = mutex_lock_interruptible(&(*kgr_dev)->mtx);
				if (ret) return ret;
				ret = kgr_isdn_net_getphones(&phone, argp);
				mutex_unlock(&(*kgr_dev)->mtx);
				return ret;
			} else
				return -EINVAL;
		case IIOCNETDNM:
			/* Delete a phone-number of a network-interface */
			if (arg) {
				if (copy_from_user(&phone, argp, sizeof(phone)))
					return -EFAULT;
				ret = mutex_lock_interruptible(&(*kgr_dev)->mtx);
				if (ret) return ret;
				ret = kgr_isdn_net_delphone(&phone);
				mutex_unlock(&(*kgr_dev)->mtx);
				return ret;
			} else
				return -EINVAL;
		case IIOCNETDIL:
			/* Force dialing of a network-interface */
			if (arg) {
				if (copy_from_user(name, argp, sizeof(name)))
					return -EFAULT;
				return kgr_isdn_net_force_dial(name);
			} else
				return -EINVAL;
#ifdef CONFIG_ISDN_PPP
		case IIOCNETALN:
			if (!arg)
				return -EINVAL;
			if (copy_from_user(name, argp, sizeof(name)))
				return -EFAULT;
			return kgr_isdn_ppp_dial_slave(name);
		case IIOCNETDLN:
			if (!arg)
				return -EINVAL;
			if (copy_from_user(name, argp, sizeof(name)))
				return -EFAULT;
			return kgr_isdn_ppp_hangup_slave(name);
#endif
		case IIOCNETHUP:
			/* Force hangup of a network-interface */
			if (!arg)
				return -EINVAL;
			if (copy_from_user(name, argp, sizeof(name)))
				return -EFAULT;
			return kgr_isdn_net_force_hangup(name);
			break;
		case IIOCSETVER:
			(*kgr_dev)->net_verbose = arg;
			printk(KERN_INFO "isdn: Verbose-Level is %d\n", (*kgr_dev)->net_verbose);
			return 0;
		case IIOCSETGST:
			if (arg)
				(*kgr_dev)->global_flags |= ISDN_GLOBAL_STOPPED;
			else
				(*kgr_dev)->global_flags &= ~ISDN_GLOBAL_STOPPED;
			printk(KERN_INFO "isdn: Global Mode %s\n",
			       ((*kgr_dev)->global_flags & ISDN_GLOBAL_STOPPED) ? "stopped" : "running");
			return 0;
		case IIOCSETBRJ:
			drvidx = -1;
			if (arg) {
				int i;
				char *p;
				if (copy_from_user(&iocts, argp,
						   sizeof(isdn_ioctl_struct)))
					return -EFAULT;
				iocts.drvid[sizeof(iocts.drvid) - 1] = 0;
				if (strlen(iocts.drvid)) {
					if ((p = strchr(iocts.drvid, ',')))
						*p = 0;
					drvidx = -1;
					for (i = 0; i < ISDN_MAX_DRIVERS; i++)
						if (!(strcmp((*kgr_dev)->drvid[i], iocts.drvid))) {
							drvidx = i;
							break;
						}
				}
			}
			if (drvidx == -1)
				return -ENODEV;
			if (iocts.arg)
				(*kgr_dev)->drv[drvidx]->flags |= DRV_FLAG_REJBUS;
			else
				(*kgr_dev)->drv[drvidx]->flags &= ~DRV_FLAG_REJBUS;
			return 0;
		case IIOCSIGPRF:
			(*kgr_dev)->profd = current;
			return 0;
			break;
		case IIOCGETPRF:
			/* Get all Modem-Profiles */
			if (arg) {
				char __user *p = argp;
				int i;

				if (!access_ok(VERIFY_WRITE, argp,
					       (ISDN_MODEM_NUMREG + ISDN_MSNLEN + ISDN_LMSNLEN)
					       * ISDN_MAX_CHANNELS))
					return -EFAULT;

				for (i = 0; i < ISDN_MAX_CHANNELS; i++) {
					if (copy_to_user(p, (*kgr_dev)->mdm.info[i].emu.profile,
							 ISDN_MODEM_NUMREG))
						return -EFAULT;
					p += ISDN_MODEM_NUMREG;
					if (copy_to_user(p, (*kgr_dev)->mdm.info[i].emu.pmsn, ISDN_MSNLEN))
						return -EFAULT;
					p += ISDN_MSNLEN;
					if (copy_to_user(p, (*kgr_dev)->mdm.info[i].emu.plmsn, ISDN_LMSNLEN))
						return -EFAULT;
					p += ISDN_LMSNLEN;
				}
				return (ISDN_MODEM_NUMREG + ISDN_MSNLEN + ISDN_LMSNLEN) * ISDN_MAX_CHANNELS;
			} else
				return -EINVAL;
			break;
		case IIOCSETPRF:
			/* Set all Modem-Profiles */
			if (arg) {
				char __user *p = argp;
				int i;

				if (!access_ok(VERIFY_READ, argp,
					       (ISDN_MODEM_NUMREG + ISDN_MSNLEN + ISDN_LMSNLEN)
					       * ISDN_MAX_CHANNELS))
					return -EFAULT;

				for (i = 0; i < ISDN_MAX_CHANNELS; i++) {
					if (copy_from_user((*kgr_dev)->mdm.info[i].emu.profile, p,
							   ISDN_MODEM_NUMREG))
						return -EFAULT;
					p += ISDN_MODEM_NUMREG;
					if (copy_from_user((*kgr_dev)->mdm.info[i].emu.plmsn, p, ISDN_LMSNLEN))
						return -EFAULT;
					p += ISDN_LMSNLEN;
					if (copy_from_user((*kgr_dev)->mdm.info[i].emu.pmsn, p, ISDN_MSNLEN))
						return -EFAULT;
					p += ISDN_MSNLEN;
				}
				return 0;
			} else
				return -EINVAL;
			break;
		case IIOCSETMAP:
		case IIOCGETMAP:
			/* Set/Get MSN->EAZ-Mapping for a driver */
			if (arg) {

				if (copy_from_user(&iocts, argp,
						   sizeof(isdn_ioctl_struct)))
					return -EFAULT;
				iocts.drvid[sizeof(iocts.drvid) - 1] = 0;
				if (strlen(iocts.drvid)) {
					drvidx = -1;
					for (i = 0; i < ISDN_MAX_DRIVERS; i++)
						if (!(strcmp((*kgr_dev)->drvid[i], iocts.drvid))) {
							drvidx = i;
							break;
						}
				} else
					drvidx = 0;
				if (drvidx == -1)
					return -ENODEV;
				if (cmd == IIOCSETMAP) {
					int loop = 1;

					p = (char __user *) iocts.arg;
					i = 0;
					while (loop) {
						int j = 0;

						while (1) {
							if (!access_ok(VERIFY_READ, p, 1))
								return -EFAULT;
							get_user(bname[j], p++);
							switch (bname[j]) {
							case '\0':
								loop = 0;
								/* Fall through */
							case ',':
								bname[j] = '\0';
								strcpy((*kgr_dev)->drv[drvidx]->msn2eaz[i], bname);
								j = ISDN_MSNLEN;
								break;
							default:
								j++;
							}
							if (j >= ISDN_MSNLEN)
								break;
						}
						if (++i > 9)
							break;
					}
				} else {
					p = (char __user *) iocts.arg;
					for (i = 0; i < 10; i++) {
						snprintf(bname, sizeof(bname), "%s%s",
							 strlen((*kgr_dev)->drv[drvidx]->msn2eaz[i]) ?
							 (*kgr_dev)->drv[drvidx]->msn2eaz[i] : "_",
							 (i < 9) ? "," : "\0");
						if (copy_to_user(p, bname, strlen(bname) + 1))
							return -EFAULT;
						p += strlen(bname);
					}
				}
				return 0;
			} else
				return -EINVAL;
		case IIOCDBGVAR:
			if (arg) {
				if (copy_to_user(argp, kgr_dev, sizeof(ulong)))
					return -EFAULT;
				return 0;
			} else
				return -EINVAL;
			break;
		default:
			if ((cmd & IIOCDRVCTL) == IIOCDRVCTL)
				cmd = ((cmd >> _IOC_NRSHIFT) & _IOC_NRMASK) & ISDN_DRVIOCTL_MASK;
			else
				return -EINVAL;
			if (arg) {
				int i;
				char *p;
				if (copy_from_user(&iocts, argp, sizeof(isdn_ioctl_struct)))
					return -EFAULT;
				iocts.drvid[sizeof(iocts.drvid) - 1] = 0;
				if (strlen(iocts.drvid)) {
					if ((p = strchr(iocts.drvid, ',')))
						*p = 0;
					drvidx = -1;
					for (i = 0; i < ISDN_MAX_DRIVERS; i++)
						if (!(strcmp((*kgr_dev)->drvid[i], iocts.drvid))) {
							drvidx = i;
							break;
						}
				} else
					drvidx = 0;
				if (drvidx == -1)
					return -ENODEV;
				if (!access_ok(VERIFY_WRITE, argp,
					       sizeof(isdn_ioctl_struct)))
					return -EFAULT;
				c.driver = drvidx;
				c.command = ISDN_CMD_IOCTL;
				c.arg = cmd;
				memcpy(c.parm.num, &iocts.arg, sizeof(ulong));
				ret = kgr_isdn_command(&c);
				memcpy(&iocts.arg, c.parm.num, sizeof(ulong));
				if (copy_to_user(argp, &iocts, sizeof(isdn_ioctl_struct)))
					return -EFAULT;
				return ret;
			} else
				return -EINVAL;
		}
	}
#ifdef CONFIG_ISDN_PPP
	if (minor <= ISDN_MINOR_PPPMAX)
		return (kgr_isdn_ppp_ioctl(minor - ISDN_MINOR_PPP, file, cmd, arg));
#endif
	return -ENODEV;

#undef name
#undef bname
#undef iocts
#undef phone
#undef cfg
}

static int kgr_patch_bsc1053150_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1053150_module_notify(struct notifier_block *nb,
					unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1053150_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1053150_module_nb = {
	.notifier_call = kgr_patch_bsc1053150_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1053150_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1053150_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1053150_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1053150_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1053150_module_nb);
}

#endif
